﻿//css_ref System.Drawing.dll;
//css_ref System.Net.Http.dll;

using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Printing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Serialization;
using Produmex.Foundation.Collections;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Sbo.Logex.Data.Utilities;

// PmxWorkflowExecutionTypes.HOOK_FLOW

namespace WorkflowScript_SalesDeliveryCreatedHookScript
{
    /// <summary>
    /// The constant parameters to configure the operation of the process
    /// </summary>
    /// <remarks>
    /// Two of these constant values may need to be changed in the future
    /// - PROCESSWEAVER_URL - the URL to the service
    /// - PROCESSWEAVER_SHIPPINGKEY - is a kind of license key, which is likely to be replaced annually
    /// </remarks>
    internal static class Definitions
	{
        public const string PROCESSWEAVER_URL = "http://10.0.0.184/RevisionSkincareAPI/API/ShippingAPI";

        // Constant parameter values for the request message
        public const string PROCESSWEAVER_SHIPPINGKEY = "00wdQYotBW0=";
        public const string PROCESSWEAVER_ERP = "SAPB1";
        public const string PROCESSWEAVER_OPERATION = "SHIP";
        public const string PROCESSWEAVER_PRINTER = "1";

        // Relevant XML tag names in response message
        public const string PROCESSWEAVER_XMLNODE_DOCURL = "DOCUMENTURL";
        public const string PROCESSWEAVER_XMLNODE_DOCNUM = "NUMBEROFDOCUMENTS";
        public const string PROCESSWEAVER_XMLNODE_MESSAGE = "MESSAGE";
        public const string PROCESSWEAVER_XMLNODE_ERROR = "ERROR";

        public const string FILENAME_PRFIX = "pmxprint_";

        public const string PAGESIZE_FOR_PRINTING = "A5";
        public const string PAGESIZE_FOR_PRINTING_STANDBY = "A4";
    }

    /// <summary>
    /// ShipRequest class for the XML request
    /// </summary>
    public class SHIPRequest
	{
        public string SHIPPINGKEY;
		public string DELIVERYNO;
        public string OPERATION;
        public string ERP;
        public string PRINTER;

        public SHIPRequest()
        {
            SHIPPINGKEY = Definitions.PROCESSWEAVER_SHIPPINGKEY;
            DELIVERYNO = "";
            OPERATION = Definitions.PROCESSWEAVER_OPERATION;
            ERP = Definitions.PROCESSWEAVER_ERP;
            PRINTER = Definitions.PROCESSWEAVER_PRINTER;
        }
	}

	/// <summary>
	/// This is a hook script, meant to be customized on a "per project" basis
	/// </summary>
	public class WorkflowScript_SalesDeliveryCreatedHookScript : WorkflowInstanceScriptBase
	{
		private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);
		private static readonly HttpClient client = new HttpClient();

        private List<string> m_documentURLs = new List<string>();
        private string m_imageFileName;
        private int m_numOfDocuments = -1;
        private string m_printer;

        private enum MessageType
        {
            WARNING,
            ERROR
        }

        private class ResponseMessage
        {
            public MessageType MsgType;
            public string Message;
        }

        private List<ResponseMessage> m_responseMessages = new List<ResponseMessage>();

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
		public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;
		public ReadOnlyBinder<DocumentRef[]> DocumentRefs;

		// Output parameters
		// <none>

		// Sub-flows
		// <none>

		/// <summary>
		/// Initializes a new instance of the <see cref="WorkflowScript_SalesDeliveryCreatedHookScript"/> class.
		/// </summary>
		/// <param name="parent">The parent.</param>
		/// <param name="factory">The factory.</param>
		public WorkflowScript_SalesDeliveryCreatedHookScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
			: base(parent, factory)
		{
		}

		#region WorkflowInstanceScriptBase Members

		protected override void Execute()
		{
			// Parameters in scope
			Session session = GetScopeParameter("Session") as Session;
			ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

			DocumentRef[] docRefs = DocumentRefs.Get();

			List<int> deliveryNums = new List<int>();

            int nonPrintableLabels = 0;

			foreach (var docRef in docRefs)
			{
				if (docRef.DocType == PmxInterfaceTypeCodes.SALES_DELIVERY)
				{
					deliveryNums.Add(docRef.DocEntry);
				}
			}

			if (deliveryNums.Count < 1)
			{
				LogError("Missing delivery note document." + ((docRefs.Length > 0 && !String.IsNullOrEmpty(docRefs[0].DocType)) ? " Received DocType: " + docRefs[0].DocType : ""));
				return;
			}

			string query = BuildQuery.GetDeliveryData(deliveryNums, sboProviderService.GetDbTool());
			DataSet ds = sboProviderService.RunView(false, null, null, query);

			PmxDictionary<int, LuidData> luidAndDeliveryNoteData = new PmxDictionary<int, LuidData>();
            string location = null;

			for (int i = 0; i < ds.Tables[0].Rows.Count; ++i)
			{
				string[] luids = SeparatorDefinition.Split(Convert.ToString(ds.Tables[0].Rows[i][PmxSalesDeliveryNoteLineTable.Columns.LogisticUnitIdentificationKey.NAME]));

				string[] locations = SeparatorDefinition.Split(Convert.ToString(ds.Tables[0].Rows[i][PmxSalesDeliveryNoteLineTable.Columns.StorageLocationCode.NAME]));

				foreach (string luidStr in luids)
				{
					try
					{
						int luid = Convert.ToInt32(luidStr);
						if (!luidAndDeliveryNoteData.ContainsKey(luid))
						{
							luidAndDeliveryNoteData.Add(
                                luid, 
								new LuidData {
                                    DeliveryNoteDocEntry = Convert.ToInt32(ds.Tables[0].Rows[i][PmxSalesDeliveryNoteLineTable.Columns.DocEntry.NAME]),
                                    SSCC = null });
						}
					}
					catch
					{
						LogError("The LUID wasn't an integer: '" + luidStr + "'");
					}
				}

                foreach (string loc in locations)
                {
                    if (!String.IsNullOrWhiteSpace(loc))
                    {
                        location = loc;
                        break;
                    }
                }
			}

			if (luidAndDeliveryNoteData.Count < 1)
			{
				LogError("There wasn't any LUIDs to the delivery notes: " + string.Join(", ", deliveryNums.Select(x => x.ToString()).ToList()));
				return;
			}

			query = BuildQuery.GetLuidDimensions(luidAndDeliveryNoteData.Keys.ToList(), sboProviderService.GetDbTool());
			ds = sboProviderService.RunView(false, null, null, query);

            SHIPRequest shipRequest = new SHIPRequest();
            bool firstSSCC = true;
            StringBuilder ssccs = new StringBuilder();

            for (int i = 0; i < ds.Tables[0].Rows.Count; ++i)
			{
				int luid = Convert.ToInt32(ds.Tables[0].Rows[i][PmxLogisticUnitIDTable.Columns.InternalKey.NAME].ToString());
				LuidData deliveryNoteData;

				if (luidAndDeliveryNoteData.TryGetValue(luid, out deliveryNoteData))
				{
                    deliveryNoteData.SSCC = ds.Tables[0].Rows[i][PmxLogisticUnitIDTable.Columns.SSCC.NAME].ToString();
                    if (!String.IsNullOrWhiteSpace(deliveryNoteData.SSCC))
                    {
                        if (firstSSCC)
                        {
                            ssccs.Append(deliveryNoteData.SSCC.Trim());
                            firstSSCC = false;
                        }
                        else
                        {
                            ssccs.Append("," + deliveryNoteData.SSCC.Trim());
                        }
                    }
                    else
                    {
                        deliveryNoteData.SSCC = null;
                    }
                }
            }

            if (ssccs.Length > 0)
            {
                shipRequest.DELIVERYNO = ssccs.ToString();

                XmlSerializer serializer = new XmlSerializer(typeof(SHIPRequest));
                XmlSerializerNamespaces ns = new XmlSerializerNamespaces();
                ns.Add("", "");

                using (StringWriter stringWriter = new Utf8StringWriter())
                {
                    using (XmlWriter writer = XmlWriter.Create(stringWriter))
                    {
                        serializer.Serialize(writer, shipRequest, ns);

                        Task<string> response = SubmitXmlRequest(stringWriter.ToString());

                        if (!String.IsNullOrEmpty(response?.Result))
                        {
                            XmlDocument xmlDoc = LoadIntoXml(response.Result);

                            if (xmlDoc != null)
                            {
                                SearchForImageUrlsInXmlResponse(xmlDoc.DocumentElement.ChildNodes);

                                if (m_documentURLs.Count > 0)
                                {
                                    RemovePreviousDownloadedFiles();

                                    if (m_numOfDocuments >= 0 && m_documentURLs.Count != m_numOfDocuments)
                                    {
                                        LogError(String.Format("The number of DOCUMENTURL tags ({0}) is not equal to the value of NUMBEROFDOCUMENTS tag ({1}). The number of SSCCs sent in request was: {2}", m_documentURLs.Count, m_numOfDocuments, luidAndDeliveryNoteData.Count));
                                    }

                                    PmxOsePrinter printer = null;

                                    sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                                        delegate (PmxDbConnection dbConn, object[] parameters)
                                        {
                                            PmxOsePrinterProvider prov = new PmxOsePrinterProvider(dbConn);
                                            printer = prov.GetPrinterForLocation(location, session.DeviceInstanceId, Definitions.PAGESIZE_FOR_PRINTING);

                                            if (printer == null && !String.IsNullOrEmpty(Definitions.PAGESIZE_FOR_PRINTING_STANDBY))
                                            {
                                                printer = prov.GetPrinterForLocation(location, session.DeviceInstanceId, Definitions.PAGESIZE_FOR_PRINTING_STANDBY);
                                            }

                                            return null;
                                        });

                                    m_printer = printer?.WindowsPrinterName;

                                    foreach (string documentUrl in m_documentURLs)
                                    {
                                        string localFileName = DownloadDocument(documentUrl);

                                        if (!String.IsNullOrEmpty(localFileName))
                                        {
                                            if (documentUrl.ToUpper().EndsWith(".PDF"))
                                            {
                                                if (!PrintPdf(localFileName))
                                                {
                                                    LogError(String.Format("Cannot print the PDF file: {0} downloaded from URL: {1}", localFileName, documentUrl));
                                                    nonPrintableLabels++;
                                                }
                                            }
                                            else
                                            {
                                                if (!PrintImage(localFileName))
                                                {
                                                    LogError(String.Format("Cannot process and print the image file: {0} downloaded from URL: {1}", localFileName, documentUrl));
                                                    nonPrintableLabels++;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            LogError("Cannot download the file from this URL: " + documentUrl);
                                            nonPrintableLabels++;
                                        }
                                    }
                                }
                                else
                                {
                                    LogError("There wasn't any image data reference in xml response.");
                                }

                                if (m_responseMessages.Count > 0)
                                {
                                    string errMessage = GetMessage();
                                    LogError(errMessage);
                                    InformTheUserAboutError(errMessage, session);
                                }
                                else if (m_documentURLs.Count == 0)
                                {
                                    InformTheUserAboutError("There was no printable label on the request.", session);
                                }
                                else if (m_documentURLs.Count == nonPrintableLabels)
                                {
                                    InformTheUserAboutError("There may have been errors in label printing.", session);
                                }
                            }
                            else
                            {
                                LogError("The response wasn't a valid xml string.");
                                InformTheUserAboutError("No label printing because of communication error (invalid XML).", session);
                            }
                        }
                        else
                        {
                            LogError("Missing response from the website: " + Definitions.PROCESSWEAVER_URL);
                            InformTheUserAboutError("No label printing because of communication error (no response).", session);
                        }
                    }
                }
            }
		}

		#endregion

		#region HELPERS

		private void LogError(string message)
		{
			s_log.Error("SalesDeliveryCreatedHookScript - " + message);
		}

        private void InformTheUserAboutError(string message, Session session)
        {
            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "MessageKey", message,
                    "NoTranslationOfMessageKey", true,
                    "IsError", true,
                    "ShowButton", true));
            WaitForMessage();
        }

        private void RemovePreviousDownloadedFiles()
        {
            try
            {
                string[] removableFiles = Directory.GetFiles(Path.GetTempPath(), Definitions.FILENAME_PRFIX + "*.*");

                foreach (string fileToRemove in removableFiles)
                {
                    File.Delete(fileToRemove);
                }
            }
            catch
            {
            }
        }

        private string DownloadDocument(string url)
        {
            try
            {
                string localFileName = Path.Combine(Path.GetTempPath(), Definitions.FILENAME_PRFIX + Path.GetFileName(url));

                using (var client = new WebClient())
                {
                    client.DownloadFile(url, localFileName);
                }

                return localFileName;
            }
            catch
            {
            }

            return null;
        }

        private async Task<string> SubmitXmlRequest(string xmlPostMessage)
		{
			var response = await client.PostAsync(Definitions.PROCESSWEAVER_URL, new StringContent(xmlPostMessage));

			return await response.Content.ReadAsStringAsync();
		}

		private XmlDocument LoadIntoXml(string xmlString)
		{
			byte[] byteArray = Encoding.ASCII.GetBytes(xmlString);
			MemoryStream stream = new MemoryStream(byteArray);
			XmlDocument retValue = new XmlDocument();

			try
			{
				retValue.Load(stream);
			}
			catch
			{
				retValue = null;
			}

			return retValue;
		}

        private void SearchForImageUrlsInXmlResponse(XmlNodeList nodeList)
        {
            foreach (XmlNode subNode in nodeList)
            {
                if (subNode.Name.ToUpper() == Definitions.PROCESSWEAVER_XMLNODE_DOCURL.ToUpper())
                {
                    if (!String.IsNullOrWhiteSpace(subNode.InnerText))
                    {
                        m_documentURLs.Add(subNode.InnerText.Trim());
                    }
                }
                else if (subNode.Name.ToUpper() == Definitions.PROCESSWEAVER_XMLNODE_DOCNUM.ToUpper())
                {
                    if (!int.TryParse(subNode.InnerText.Trim(), out m_numOfDocuments))
                    {
                        m_numOfDocuments = 0;
                    }
                }
                else if (subNode.Name.ToUpper() == Definitions.PROCESSWEAVER_XMLNODE_MESSAGE.ToUpper())
                {
                    m_responseMessages.Add(new ResponseMessage() { MsgType = MessageType.WARNING, Message = subNode.InnerText.Trim() });
                }
                else if (subNode.Name.ToUpper() == Definitions.PROCESSWEAVER_XMLNODE_ERROR.ToUpper())
                {
                    m_responseMessages.Add(new ResponseMessage() { MsgType = MessageType.ERROR, Message = subNode.InnerText.Trim() });
                }
                else if (subNode.ChildNodes.Count > 0)
                {
                    SearchForImageUrlsInXmlResponse(subNode.ChildNodes);
                }
            }
        }

        private string GetMessage()
        {
            StringBuilder strBuilder = new StringBuilder();

            foreach (var msgAndType in m_responseMessages)
            {
                if (strBuilder.Length > 0)
                {
                    strBuilder.Append(Environment.NewLine);
                }

                strBuilder.AppendFormat("{0}: {1}" , Enum.GetName(typeof(MessageType), msgAndType.MsgType), msgAndType.Message);
            }

            return strBuilder.ToString();
        }

        private bool PrintImage(string imageFileName)
        {
            try
            {
                PrintDocument pd = new PrintDocument();

                if (!String.IsNullOrEmpty(m_printer))
                {
                    pd.PrinterSettings.PrinterName = m_printer;
                }

                m_imageFileName = imageFileName;
                pd.PrintPage += PrintPage;
                pd.Print();
            }
            catch
            {
                return false;
            }

            return true;
        }

        private void PrintPage(object o, PrintPageEventArgs e)
        {
            Image image = Image.FromFile(m_imageFileName);

            float newWidth = image.Width * 100 / image.HorizontalResolution;
            float newHeight = image.Height * 100 / image.VerticalResolution;

            float widthFactor = newWidth / e.MarginBounds.Width;
            float heightFactor = newHeight / e.MarginBounds.Height;

            if (widthFactor > 1 | heightFactor > 1)
            {
                if (widthFactor > heightFactor)
                {
                    newWidth = newWidth / widthFactor;
                    newHeight = newHeight / widthFactor;
                }
                else
                {
                    newWidth = newWidth / heightFactor;
                    newHeight = newHeight / heightFactor;
                }
            }

            e.Graphics.DrawImage(image, 0, 0, (int)newWidth, (int)newHeight);
        }

        private bool PrintPdf(string pdfFileName)
        {
            try
            {
                ProcessStartInfo info;

                if (!String.IsNullOrEmpty(m_printer))
                {
                    info = new ProcessStartInfo(pdfFileName, "\"" + m_printer + "\"");
                }
                else
                {
                    info = new ProcessStartInfo(pdfFileName);
                }

                info.Verb = "printto";
                info.CreateNoWindow = true;
                info.WindowStyle = ProcessWindowStyle.Hidden;

                Process p = new Process();
                p.StartInfo = info;
                p.Start();

                p.WaitForInputIdle();

                System.Threading.Thread.Sleep(3000);
                if (false == p.CloseMainWindow())
                    p.Kill();
            }
            catch
            {
                return false;
            }

            return true;
        }

        #endregion
    }

    public class LuidData
    {
        public int DeliveryNoteDocEntry;
        public string SSCC;
    }

    public class Utf8StringWriter : StringWriter
    {
        public override Encoding Encoding { get { return Encoding.UTF8; } }
    }

    #region BUILD QUERY

    internal static class BuildQuery
	{
		public static string GetDeliveryData(List<int> deliveryNums, DbTool dbTool)
		{
			DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

			strBuilder.
				Select.
					EscTableColumnName(PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.DocEntry.NAME).Comma.
					EscTableColumnName(PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.StorageLocationCode.NAME).Comma.
					EscTableColumnName(PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.LogisticUnitIdentificationKey.NAME).CRLF.
				From.
					EscTableName(PmxSalesDeliveryNoteLineTable.NAME).CRLF.
				Where.
					EscTableColumnName(PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.DocEntry.NAME).In.
					AppendFormat("( {0} ) ", SqlCommandHelper.FormatColumnValue(deliveryNums, dbTool.CultureInfo));

			return strBuilder.ToString();
		}

		public static string GetLuidDimensions(List<int> luids, DbTool dbTool)
		{
			DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

			strBuilder.
				Select.
                    EscTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME).Comma.
                    EscTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME).CRLF.
				From.
					EscTableName(PmxLogisticUnitIDTable.NAME).CRLF.
				Where.
					EscTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME).In.
					AppendFormat("( {0} ) ", SqlCommandHelper.FormatColumnValue(luids, dbTool.CultureInfo));

			return strBuilder.ToString();
		}
	}

	#endregion
}
